/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file CommunicationManagerImpl.cpp
 *
 * @date 2011-10-26
 * @author Philipp Kegel
 */

#include "CommunicationManagerImpl.h"

#include "ComputeNodeImpl.h"
#include "HostImpl.h"
#include "ProcessImpl.h"

#include "comm/CLMessageDispatcher.h"

#include "message/ContextErrorMessage.h"
#include "message/DeviceIDsResponse.h"
#include "message/DeviceInfosResponse.h"
#include "message/GetDeviceIDs.h"
#include "message/GetDeviceInfo.h"
#include "message/ProcessMessage.h"
#include "message/ProgramBuildMessage.h"

#include <dclgcf/message/BuildProgram.h>
#include <dclgcf/message/CreateCommandQueue.h>
#include <dclgcf/message/CreateBuffer.h>
#include <dclgcf/message/CreateContext.h>
#include <dclgcf/message/CreateEvent.h>
#include <dclgcf/message/CreateKernel.h>
#include <dclgcf/message/CreateKernelsInProgram.h>
#include <dclgcf/message/CreateProgramWithSource.h>
#include <dclgcf/message/DeleteMemory.h>
#include <dclgcf/message/DeleteCommandQueue.h>
#include <dclgcf/message/DeleteContext.h>
#include <dclgcf/message/DeleteEvent.h>
#include <dclgcf/message/DeleteKernel.h>
#include <dclgcf/message/DeleteProgram.h>
#include <dclgcf/message/EnqueueBarrier.h>
#include <dclgcf/message/EnqueueCopyBuffer.h>
#include <dclgcf/message/EnqueueMapBuffer.h>
#include <dclgcf/message/EnqueueMarker.h>
#include <dclgcf/message/EnqueueNDRangeKernel.h>
#include <dclgcf/message/EnqueueReadBuffer.h>
#include <dclgcf/message/EnqueueUnmapMemObject.h>
#include <dclgcf/message/EnqueueWaitForEvents.h>
#include <dclgcf/message/EnqueueWriteBuffer.h>
#include <dclgcf/message/ErrorResponse.h>
#include <dclgcf/message/EventProfilingInfosResponse.h>
#include <dclgcf/message/EventSynchronizationMessage.h>
#include <dclgcf/message/CommandMessage.h>
#include <dclgcf/message/FinishRequest.h>
#include <dclgcf/message/FlushRequest.h>
#include <dclgcf/message/GetEventProfilingInfos.h>
#include <dclgcf/message/GetKernelInfo.h>
#include <dclgcf/message/InfoResponse.h>
#include <dclgcf/message/Request.h>
#include <dclgcf/message/Response.h>
#include <dclgcf/message/SetKernelArg.h>

#include <dcl/ConnectionListener.h>
#include <dcl/DCLException.h>
#include <dcl/DCLTypes.h>

#include <dcl/util/Logger.h>

#include <gcf/CallbackPolicy.h>
#include <gcf/CommunicationManager.h>
#include <gcf/CommunicationManagerProperties.h>
#include <gcf/ConnectionListener.h>
#include <gcf/GCFTypedefs.h>
#include <gcf/Process.h>
#include <gcf/Properties.h>
#include <gcf/ReplicateMappings.h>
#include <gcf/Serializable.h>
#include <gcf/SerializationModule.h>
#include <gcf/StreamCommunication.h>

#include <gcf/link/SocketOptions.h>

#include <gcf/transport/DataStream.h>
#include <gcf/transport/IODataStream.h>

#include <boost/algorithm/string/trim.hpp>

#include <cassert>
#include <iterator>
#include <memory>
#include <mutex>
#include <ostream>
#include <string>
#include <utility>
#include <vector>

namespace dclgcf {

/******************************************************************************
 *
 * Communication manager base implementation
 *
 ******************************************************************************/

const std::chrono::seconds CommunicationManagerImpl::DEFAULT_CONNECTION_TIMEOUT = std::chrono::seconds(3);

void CommunicationManagerImpl::resolveURL(
		const std::string& url,
		std::string& hostName,
		gcf_uint16& port) {
	size_t delim = url.find(":");

	hostName = url.substr(0, delim);
	boost::trim(hostName); // remove leading and trailing white spaces
	if (delim != std::string::npos) {
		std::istringstream stream(url.substr(delim));
		gcf_uint16 defaultPort = port;

		stream.ignore(); // skip delimiter
		if ((stream >> port).fail()) {
			port = defaultPort;
		}
	}
}

CommunicationManagerImpl::CommunicationManagerImpl(
		const gcf::CommunicationManagerProperties& commProps) {
	try {
		gcf::Properties props;

		props.setProperty(&commProps);
		_communicationManager.reset(new gcf::CommunicationManager(props));
	} catch (const gcf::GCFError& err) {
		throw dcl::ConnectionException(err.message);
	}

	/*
	 * Register all dOpenCL message types
	 * Thus, invalid messages can also be de-/serialized by processes (host,
	 * compute node, resource manager) such that they can report a protocol
	 * error to the application.
	 */
	auto& sm = _communicationManager->getSerializationModule();

	/*
	 * Register CL command requests
	 */
	message::Request::registerSerializable(sm);
	message::BuildProgram::registerSerializable(sm);
	message::CreateBuffer::registerSerializable(sm);
	message::CreateCommandQueue::registerSerializable(sm);
	message::CreateContext::registerSerializable(sm);
	message::CreateEvent::registerSerializable(sm);
	message::CreateKernel::registerSerializable(sm);
	message::CreateKernelsInProgram::registerSerializable(sm);
	message::CreateProgramWithSource::registerSerializable(sm);
	message::DeleteMemory::registerSerializable(sm);
	message::DeleteContext::registerSerializable(sm);
	message::DeleteCommandQueue::registerSerializable(sm);
	message::DeleteEvent::registerSerializable(sm);
	message::DeleteKernel::registerSerializable(sm);
	message::DeleteProgram::registerSerializable(sm);
	message::EnqueueBarrier::registerSerializable(sm);
	message::EnqueueCopyBuffer::registerSerializable(sm);
	message::EnqueueMarker::registerSerializable(sm);
    message::EnqueueMapBuffer::registerSerializable(sm);
	message::EnqueueNDRangeKernel::registerSerializable(sm);
	message::EnqueueReadBuffer::registerSerializable(sm);
    message::EnqueueUnmapMemObject::registerSerializable(sm);
	message::EnqueueWaitForEvents::registerSerializable(sm);
	message::EnqueueWriteBuffer::registerSerializable(sm);
	message::FinishRequest::registerSerializable(sm);
	message::FlushRequest::registerSerializable(sm);
	message::GetDeviceIDs::registerSerializable(sm);
	message::GetDeviceInfo::registerSerializable(sm);
	message::GetEventProfilingInfos::registerSerializable(sm);
	message::GetKernelInfo::registerSerializable(sm);
    message::GetKernelWorkGroupInfo::registerSerializable(sm);
	message::SetKernelArg::registerSerializable(sm);
	message::SetKernelArgBinary::registerSerializable(sm);
	message::SetKernelArgMemObject::registerSerializable(sm);

	/*
	 * Register CL command responses
	 */
	message::Response::registerSerializable(sm);
	message::DeviceIDsResponse::registerSerializable(sm);
	message::DeviceInfosResponse::registerSerializable(sm);
	message::ErrorResponse::registerSerializable(sm);
	message::EventProfilingInfosReponse::registerSerializable(sm);
	message::InfoResponse::registerSerializable(sm);

	/*
	 * Register CL notifications
	 */
	message::CommandExecutionStatusChangedMessage::registerSerializable(sm);
	message::ContextErrorMessage::registerSerializable(sm);
	message::ProgramBuildMessage::registerSerializable(sm);
	message::EventSynchronizationMessage::registerSerializable(sm);

	/*
	 * Register process status messages
	 */
	message::HostIdentificationMessage::registerSerializable(sm);
	message::ComputeNodeIdentificationMessage::registerSerializable(sm);
	message::ConnectionApprovedMessage::registerSerializable(sm);
}

CommunicationManagerImpl::~CommunicationManagerImpl() {
}

void CommunicationManagerImpl::start() {
    std::vector<GCFTypeID_t> processMessageTypes;

    /* create list of process status message types*/
    processMessageTypes.push_back(message::HostIdentificationMessage::TYPE);
    processMessageTypes.push_back(message::ComputeNodeIdentificationMessage::TYPE);
    processMessageTypes.push_back(message::ConnectionApprovedMessage::TYPE);

	_communicationManager->addMessageListener(*this,
	        gcf::CallbackPolicy(processMessageTypes, false));

	/* Set buffered to false to immediately register new sessions */
    _communicationManager->addConnectionListener(*this, false);
    _communicationManager->getStreamCommunication().addDataStreamListener(*this);

	_communicationManager->activateCommunication();
}

void CommunicationManagerImpl::stop() {
    _communicationManager->getStreamCommunication().removeDataStreamListener(*this);
	_communicationManager->removeConnectionListener(*this);

	_communicationManager->removeMessageListener(*this);

	_communicationManager->shutdownCommunication();
}

ProcessImpl * CommunicationManagerImpl::getProcess(
        const gcf::Process& process) const {
    return getComputeNode(process);
}

ComputeNodeImpl * CommunicationManagerImpl::getComputeNode(
        const gcf::Process& process) const {
    std::lock_guard<std::recursive_mutex> lock(_connectionsMutex);

    /* search compute node list */
    auto i = _computeNodes.find(&process);
    if (i != std::end(_computeNodes)) return i->second;

    return nullptr;
}

#if 0
/*
 * Assume that all compute nodes have been introduced statically by a configuration file.
 */
void CommunicationManagerImpl::getComputeNodes(
        const std::vector<dcl_object_id>& computeNodeIds,
        std::vector<ComputeNodeImpl *>& computeNodes) {
    std::lock_guard<std::mutex> lock(_connectionsMutex);

    for (auto computeNodeId : computeNodeIds) {
        auto i = _computeNodes.find(computeNodeId);
        if (i == std::end(_computeNodes)) { // compute node not found
            throw dcl::ConnectionException("Invalid compute node ID");
        }
        computeNodes.push_back(i->second);
    }
}
#endif

/*
 * GCF connection listener API
 */

bool CommunicationManagerImpl::approveConnection(
		const std::string& subjectName) {
	/* This method is not only called for incoming process connections but also
	 * for outgoing connections. */
	return true;
}

void CommunicationManagerImpl::connectionEstablished(
		const gcf::Process& process, const std::string& subjectName) {
	/* TODO Distinguish incoming (host only) and outgoing (host, compute node, device manager) connections
	 * This method is not only called for incoming process connections but also
	 * for outgoing connections. */

	/* TODO Add incoming connections to list of pending connections
	 * Connections will not be used until an identification has been send by
	 * the remote process. To avoid idle connections, pending connections should
	 * be closed after a certain timeout. */
}

void CommunicationManagerImpl::connectionLost(const gcf::Process& process) {
	if (hostConnectionLost(process)) return;
	if (computeNodeConnectionLost(process)) return;

	dcl::util::Logger << dcl::util::Warning <<
			"Lost connection to unknown process" << std::endl;
}

gcf::ReplicateMappings CommunicationManagerImpl::incomingConnectionTransfer(
		const gcf::Process& host,
		const gcf::Process& computeNode,
		gcf::Serializable *transferMessage,
		std::vector<gcf::Serializable *>& connectedEntities,
		std::vector<gcf::Serializable *>& newEntities) {
	/* TODO Handle incoming connection transfers */
	return gcf::ReplicateMappings();
}

/*
 * GCF data stream listener API
 */

bool CommunicationManagerImpl::incomingDataStream(
        gcf::IODataStream &dataStream) {
    auto process = dataStream.getRemoteProcess();
    if (process) {
        /* The connections list must be locked, such that a node is not
         * destroyed while its associated data stream is changed */
        std::lock_guard<std::recursive_mutex> lock(_connectionsMutex);

        /* lookup process' associated dOpenCL node */
        auto node = getProcess(*process);
        if (node) {
            node->setDataStream(&dataStream);
            /* TODO Log node type ('host' or 'compute node') */
            dcl::util::Logger << dcl::util::Info
                    << "Incoming data stream from node '" << node->url() << '\''
                    << std::endl;

            return true;
        } else {
            dcl::util::Logger << dcl::util::Warning
                    << "Incoming data stream from unknown node" << std::endl;
        }
    } else {
        dcl::util::Logger << dcl::util::Error
                << "No process for data stream" << std::endl;
    }

    return false;
}

/* FIXME Ensure that closed data streams are detected
 * This method is not called when a data stream is closed by the remote endpoint
 * while no data transfer is pending (see implementation in
 * gcf::DataStreamProcessor). Hence, ProcessImpl::setDataStream(nullptr) is
 * called when the process connection is lost. */
void CommunicationManagerImpl::closedDataStream(
        gcf::DataStream& dataStream) {
    auto process = dataStream.getRemoteProcess();
    if (process) {
        /* The connections list must be locked, such that a node is not
         * destroyed while its associated data stream is changed */
        std::lock_guard<std::recursive_mutex> lock(_connectionsMutex);

        /* lookup process' associated dOpenCL node */
        auto node = getProcess(*process);
        if (node) { // node found
            node->setDataStream(nullptr);
            /* TODO Log node type ('host' or 'compute node') */
            dcl::util::Logger << dcl::util::Info
                    << "Closed data stream from node '" << node->url() << '\''
                    << std::endl;
        } else {
            dcl::util::Logger << dcl::util::Warning
                    << "Closed data stream from unknown node" << std::endl;
        }
    } else {
        dcl::util::Logger << dcl::util::Error
                << "No process for data stream" << std::endl;
    }
}

/*
 * GCF message listener API
 */

void CommunicationManagerImpl::dispatch(
		gcf::Serializable& message, const gcf::Process& process) {
	switch (message.getType()) {

	/*
	 * Process identification message
	 */
	case message::HostIdentificationMessage::TYPE: // process is host
		hostConnectionEstablished(process);
		break;
	case message::ComputeNodeIdentificationMessage::TYPE: // process is compute node
		computeNodeConnectionEstablished(process);
		break;

	/*
	 * Process status messages
	 */
	case message::ConnectionApprovedMessage::TYPE:
		connectionApproved(process);
		break;

	default: /* invalid message */
		dcl::util::Logger << dcl::util::Error
				<< "Received message of unknown type" << std::endl;
		break;
	}
}

void CommunicationManagerImpl::hostConnectionEstablished(const gcf::Process&) {
	/* ignore process */
	dcl::util::Logger << dcl::util::Warning
			<< "Ignoring incoming host connection" << std::endl;
}

void CommunicationManagerImpl::computeNodeConnectionEstablished(const gcf::Process&) {
	/* ignore process */
	dcl::util::Logger << dcl::util::Warning
			<< "Ignoring incoming compute node connection" << std::endl;
}

bool CommunicationManagerImpl::hostConnectionLost(const gcf::Process&) {
	return false;
}

bool CommunicationManagerImpl::computeNodeConnectionLost(const gcf::Process&) {
	return false;
}

void CommunicationManagerImpl::connectionApproved(
        const gcf::Process& process) const {
    auto computeNode = getComputeNode(process);
    if (computeNode) {
        computeNode->setConnectionStatus(ComputeNodeImpl::ConnectionStatus::CONNECTED); // update compute node connection status
    } else {
        dcl::util::Logger << dcl::util::Error
                << "Received connection approval from unknown process" << std::endl;
    }
}

} /* namespace dclgcf */
