/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file ComputeNodeCommunicationManagerImpl.cpp
 *
 * @date 2011-10-26
 * @author Philipp Kegel
 */

#include "ComputeNodeCommunicationManagerImpl.h"

#include "ComputeNodeImpl.h"
#include "HostImpl.h"
#include "ProcessImpl.h"
#include "SmartCLObjectRegistry.h"

#include "comm/CLMessageDispatcher.h"
#include "comm/CLRequestDispatcher.h"

#include "message/ProcessMessage.h"

#include <dcl/ComputeNode.h>
#include <dcl/ConnectionListener.h>
#include <dcl/Daemon.h>
#include <dcl/DCLException.h>
#include <dcl/DCLTypes.h>

#include <dcl/util/Logger.h>

#include <gcf/CommunicationManager.h>
#include <gcf/CommunicationManagerProperties.h>
#include <gcf/ConnectionListener.h>
#include <gcf/Process.h>
#include <gcf/ProcessLookup.h>
#include <gcf/Properties.h>
#include <gcf/ReplicateMappings.h>
#include <gcf/Serializable.h>
#include <gcf/SerializationModule.h>

#include <cassert>
#include <condition_variable>
#include <iterator>
#include <memory>
#include <mutex>
#include <ostream>
#include <set>
#include <string>
#include <utility>
#include <vector>

namespace dclgcf {

/******************************************************************************
 *
 * Compute node communication manager implementation
 *
 ******************************************************************************/

ComputeNodeCommunicationManagerImpl::ComputeNodeCommunicationManagerImpl(
		const gcf::CommunicationManagerProperties& commProps) :
		CommunicationManagerImpl(commProps) {
	dcl::util::Logger << dcl::util::Info
			<< "Bound to " << commProps.hostname << ':' << commProps.port
			<< std::endl;

	_clMessageDispatcher.reset(new comm::CLHostMessageDispatcher(*this));
	_clRequestDispatcher.reset(new comm::CLRequestDispatcher(*this));
}

ComputeNodeCommunicationManagerImpl::~ComputeNodeCommunicationManagerImpl() {
}

void ComputeNodeCommunicationManagerImpl::start() {
    _communicationManager->addMessageListener(*_clMessageDispatcher,
            comm::CLMessageDispatcher::callbackPolicy());
	_communicationManager->addMessageListener(*_clRequestDispatcher,
	        comm::CLRequestDispatcher::callbackPolicy());

	CommunicationManagerImpl::start();
}

void ComputeNodeCommunicationManagerImpl::stop() {
    _communicationManager->removeMessageListener(*_clMessageDispatcher);
	_communicationManager->removeMessageListener(*_clRequestDispatcher);

	CommunicationManagerImpl::stop();
}

SmartCLObjectRegistry& ComputeNodeCommunicationManagerImpl::objectRegistry() {
    return _objectRegistry;
}

void ComputeNodeCommunicationManagerImpl::createComputeNodes(
        const std::vector<std::string>& urls,
        std::vector<ComputeNodeImpl *>& computeNodes) {
    std::set<const gcf::Process *> processes;

    /*
     * Create process list
     */
    for (auto url : urls) {
        std::string host;
        gcf_uint16 port = DEFAULT_COMPUTE_NODE_PORT;

        resolveURL(url, host, port);
        assert(!host.empty());

        /* WARNING: Broken GCF API design
         * For the following line to work correctly,
         * USE_PROCESSLOOKUP_SINGLETON must be (un)defined, if it has been
         * (un)defined during the compilation of GCF */
        gcf::Process& process = _communicationManager->getProcessLookup().createProcess(host, port);
        bool inserted = processes.insert(&process).second;
        assert(inserted); // URL must only be specified once
    }

    createComputeNodes(processes, computeNodes);
}

void ComputeNodeCommunicationManagerImpl::createComputeNodes(
        const std::set<const gcf::Process *>& processes,
        std::vector<ComputeNodeImpl *>& computeNodes) {
    std::lock_guard<std::recursive_mutex> lock(_connectionsMutex);

    computeNodes.clear();
    computeNodes.reserve(processes.size());

    /*
     * (Lookup or) create compute nodes and add to list
     */
    for (auto process : processes) {
        assert(process != nullptr); // process must not be NULL
        /* WARNING: The equals operator can throw an assertion (see implementation
         * of gcf::Process), if the local and the remote process have not been
         * connected yet. Therefore, the local process must be properly initialized. */
        if (_communicationManager->getOwnProcess().equals(process)) {
            /* TODO Return reference to local compute node instance */
            continue;
        }

        auto i = _computeNodes.find(process);
        if (i == std::end(_computeNodes)) {
            auto computeNode = new ComputeNodeImpl(
                    *_communicationManager, *process);

            /* Process will not have a valid process ID before it is connected.
             * Hence the process ID cannot be used to identify a process. */
            _computeNodes.insert(std::make_pair(process, computeNode));
            computeNodes.push_back(computeNode);
        } else {
            computeNodes.push_back(i->second);
        }
    }
}

void ComputeNodeCommunicationManagerImpl::destroyComputeNode(
        ComputeNodeImpl *computeNode) {
    if (computeNode) {
        const gcf::Process *process = &computeNode->process();

        computeNode->disconnect();

        {
            std::lock_guard<std::recursive_mutex> lock(_connectionsMutex);
            _computeNodes.erase(process);
        }

        delete computeNode;
        _communicationManager->getProcessLookup().freeProcess(process);
    } else {
        throw dcl::InvalidArgument(DCL_INVALID_NODE);
    }
}

void ComputeNodeCommunicationManagerImpl::connectComputeNodes(
        const std::vector<ComputeNodeImpl *>& computeNodes) {
    /* Compute node that will connect to this process */
    std::set<ComputeNodeImpl *> activeComputeNodes;
    /* Compute nodes that this process will connect to */
    std::set<ComputeNodeImpl *> passiveComputeNodes;

    /* A collective node-to-node connection operation:
     * A connection is be initiated by exactly one of the involved compute
     * nodes. Therefore, compute nodes must agree on which ones actively connect
     * while others passively accept connections.
     * For this purpose, an order of compute nodes is derived from the order of
     * GCF processes. Based on this order, computes nodes either have a 'lower'
     * or 'higher' rank than others.
     * A compute node only actively connects to other compute nodes that have a
     * lower rank, and passively wait for connections from compute nodes with
     * higher ranks. */

    /*
     * Determine active and passive compute nodes
     */
    for (auto computeNode : computeNodes) {
        assert(computeNode != nullptr); // compute node must not be NULL
        const auto& process = (computeNode)->process();

        /* WARNING: The equals operator can throw an assertion (see implementation
         * of gcf::Process), if the local and the remote process have not been
         * connected yet. Therefore, the local process must be properly initialized. */
        if (_communicationManager->getOwnProcess().equals(&process)) {
            /* a compute node cannot connect to itself */
            continue;
        }

        /* WARNING Do not use operator< to compare processes */
        if (process.getProcessId() < _communicationManager->getOwnProcess().getProcessId()) {
            /* actively connect to passive compute nodes */
            passiveComputeNodes.insert(computeNode);
        } else {
            /* await connection from active compute nodes */
            activeComputeNodes.insert(computeNode);
        }
    }

    /* Connect to passive compute nodes (parallelized operation) */
    ComputeNodeImpl::connect(
            std::vector<ComputeNodeImpl *>(std::begin(passiveComputeNodes), std::end(passiveComputeNodes)),
            ProcessImpl::Type::COMPUTE_NODE);

    /* Await connection of active compute nodes */
    ComputeNodeImpl::awaitConnection(
            std::vector<ComputeNodeImpl *>(std::begin(activeComputeNodes), std::end(activeComputeNodes)));
}

void ComputeNodeCommunicationManagerImpl::setDaemon(dcl::Daemon *daemon) {
    if (_daemon) {
        /* TODO Deregister devices of old daemon */
    }

    _daemon = daemon;

    if (_daemon) {
        std::vector<dcl::Device *> devices;

        /* Register devices */
        _daemon->getDevices(devices);
        for (auto device : devices) {
            _objectRegistry.bind(device->getId(), device);
        }
    }
}

dcl::Daemon * ComputeNodeCommunicationManagerImpl::getDaemon() const {
    return _daemon;
}

bool ComputeNodeCommunicationManagerImpl::addConnectionListener(
		dcl::ConnectionListener& listener) {
	std::lock_guard<std::mutex> lock(_connectionListenersMutex);
	return (_connectionListeners.insert(&listener)).second;
}

bool ComputeNodeCommunicationManagerImpl::removeConnectionListener(
		dcl::ConnectionListener& listener) {
	std::lock_guard<std::mutex> lock(_connectionListenersMutex);
	return (_connectionListeners.erase(&listener) == 1);
}

ProcessImpl * ComputeNodeCommunicationManagerImpl::getProcess(
        const gcf::Process& process) const {
    std::lock_guard<std::recursive_mutex> lock(_connectionsMutex);

    /* Do not use ComputeNodeCommunicationManagerImpl::getHost and
     * CommunicationManagerImpl::getComputeNode to search for node in order to
     * avoid multiple locking operations on _connectionsMutex */

    /* search host list */
    auto i = _hosts.find(&process);
    if (i != std::end(_hosts)) return i->second;

    /* search compute node list */
    auto j = _computeNodes.find(&process);
    if (j != std::end(_computeNodes)) return j->second;

    return nullptr;
}

HostImpl * ComputeNodeCommunicationManagerImpl::getHost(
        const gcf::Process& process) const {
    std::lock_guard<std::recursive_mutex> lock(_connectionsMutex);

    /* search host list */
    auto i = _hosts.find(&process);
    if (i != std::end(_hosts)) return i->second;

    return nullptr;
}

void ComputeNodeCommunicationManagerImpl::hostConnectionEstablished(
		const gcf::Process& process) {
    assert(getComputeNode(process) == nullptr); // process must not be associated with a compute node yet
	assert(getHost(process) == nullptr); // host must not have been created yet

    HostImpl *host = new HostImpl(*_communicationManager, process);
    bool accepted = false;

	/*
	 * Notify connection listeners
	 */
	{
		std::lock_guard<std::mutex> lock(_connectionListenersMutex);
		for (auto listener : _connectionListeners) {
		    /* WARNING: do not change expression order, otherwise short-cut
		     * evaluation might skip calling the listener */
			accepted = listener->connected(*host) || accepted;
		}
	}

	if (accepted) {
        /* add host to list */
	    {
            std::lock_guard<std::recursive_mutex> lock(_connectionsMutex);
            bool inserted = _hosts.insert(std::make_pair(&process, host)).second;
            assert(inserted && "Could not add host to list"); // assert insertion of host into list
	    }

        /* send connection approval */
        message::ConnectionApprovedMessage message;
        _communicationManager->send(process, message);

        dcl::util::Logger << dcl::util::Info
                << "Accepted connection from host " << host->url() << std::endl;
	} else {
	    /* TODO Send rejection message if incoming connection has been rejected */

	    delete host;

        dcl::util::Logger << dcl::util::Warning
                << "Rejecting connection from host " << host->url() << std::endl;
	}
}

void ComputeNodeCommunicationManagerImpl::computeNodeConnectionEstablished(
		const gcf::Process& process) {
    assert(getHost(process) == nullptr); // process must not be associated with a host yet

    ComputeNodeImpl *computeNode = getComputeNode(process);;
    bool accepted = false;

    if (computeNode) {
        /* update compute node's connection status */
        computeNode->setConnectionStatus(ProcessImpl::ConnectionStatus::CONNECTED);
    } else {
        computeNode = new ComputeNodeImpl(*_communicationManager, process);
    }

	/*
	 * Notify connection listeners
	 */
	{
		std::lock_guard<std::mutex> lock(_connectionListenersMutex);
		for (auto listener : _connectionListeners) {
            /* WARNING: do not change expression order, otherwise short-cut
             * evaluation might skip calling the listener */
			accepted = listener->connected(*computeNode) || accepted;
		}
	}

    if (accepted) {
        /* add compute node to list;
         * insertion may fail if compute node is re-connected */
        {
            std::lock_guard<std::recursive_mutex> lock(_connectionsMutex);
            _computeNodes.insert(std::make_pair(&process, computeNode));
        }

        /* send connection approval */
        message::ConnectionApprovedMessage message;
        _communicationManager->send(process, message);

        dcl::util::Logger << dcl::util::Info
                << "Accepted connection from compute node " << computeNode->url() << std::endl;
    } else {
        /* remove compute node from list */
        {
            std::lock_guard<std::recursive_mutex> lock(_connectionsMutex);
            _computeNodes.erase(&process);
        }

        /* TODO Send rejection message if incoming connection has been rejected */

        delete computeNode;

        dcl::util::Logger << dcl::util::Warning
                << "Rejected connection from compute node " << computeNode->url() << std::endl;
    }
}

bool ComputeNodeCommunicationManagerImpl::hostConnectionLost(
        const gcf::Process& process) {
    HostImpl *host = getHost(process);
    if (host) {
        dcl::util::Logger << dcl::util::Warning
                << "Lost connection to host '" << host->url() << '\''
                << std::endl;

        host->setConnectionStatus(HostImpl::ConnectionStatus::DISCONNECTED); // update host connection status
        /* FIXME Enable GCF DataStreamProcessor to detect closed streams
         * Deleting the data stream is a work-around to ensure that a closed
         * stream is removed from the GCF DataStreamProcessor. Otherwise, the
         * stream will make the DataStreamProcessor to spin in its data receipt
         * procedure, as a closed socket is indicated as incoming data. */
        host->setDataStream(nullptr);

        /*
         * Notify connection listeners
         */
        {
            std::lock_guard<std::mutex> lock(_connectionListenersMutex);
            for (auto listener : _connectionListeners) {
                listener->disconnected(*host);
            }
        }
    }

    return host != nullptr;
}

bool ComputeNodeCommunicationManagerImpl::computeNodeConnectionLost(
        const gcf::Process& process) {
    ComputeNodeImpl *computeNode = getComputeNode(process);
    if (computeNode) {
        dcl::util::Logger << dcl::util::Warning
                << "Lost connection to compute node '" << computeNode->url() << '\''
                << std::endl;

        /* TODO Attempt to re-established failed connection transparently */

        computeNode->setConnectionStatus(ComputeNodeImpl::ConnectionStatus::DISCONNECTED);
        /* FIXME Enable GCF DataStreamProcessor to detect closed streams
         * Deleting the data stream is a work-around to ensure that a closed
         * stream is removed from the GCF DataStreamProcessor. Otherwise, the
         * stream will make the DataStreamProcessor spin in its data receipt
         * procedure, as a closed socket is indicated as incoming data. */
        computeNode->setDataStream(nullptr);

        /*
         * Notify connection listeners
         */
        {
            std::lock_guard<std::mutex> lock(_connectionListenersMutex);
            for (auto listener : _connectionListeners) {
                listener->disconnected(*computeNode);
            }
        }
    }

    return computeNode != nullptr;
}

} /* namespace dclgcf */
