/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file HostCommunicationManagerImpl.cpp
 *
 * @date 2011-10-26
 * @author Philipp Kegel
 */

#include "HostCommunicationManagerImpl.h"

#include "ComputeNodeImpl.h"
#include "HostImpl.h"

#include "comm/CLMessageDispatcher.h"
#include "comm/CLResponseDispatcher.h"

#include <dcl/CLObjectRegistry.h>
#include <dcl/ComputeNode.h>
#include <dcl/ContextListener.h>
#include <dcl/DCLException.h>
#include <dcl/DCLTypes.h>

#include <dcl/util/Logger.h>

#include <gcf/CommunicationManager.h>
#include <gcf/CommunicationManagerProperties.h>
#include <gcf/Process.h>
#include <gcf/ProcessLookup.h>
#include <gcf/Properties.h>
#include <gcf/ReplicateMappings.h>
#include <gcf/Serializable.h>
#include <gcf/SerializationModule.h>

#include <cassert>
#include <iterator>
#include <memory>
#include <mutex>
#include <set>
#include <string>
#include <utility>
#include <vector>

namespace dclgcf {

/******************************************************************************
 *
 * Host communication manager implementation
 *
 ******************************************************************************/

HostCommunicationManagerImpl::HostCommunicationManagerImpl() :
		CommunicationManagerImpl(
				/* do not bind this process to any address */
				gcf::CommunicationManagerProperties("", 0, false)) {
    _clMessageDispatcher.reset(new comm::CLComputeNodeMessageDispatcher(
            *this, _objectRegistry));
    _clResponseDispatcher.reset(new comm::CLResponseDispatcher(*this));
}

HostCommunicationManagerImpl::~HostCommunicationManagerImpl() {
}

void HostCommunicationManagerImpl::start() {
    _communicationManager->addMessageListener(*_clMessageDispatcher,
            comm::CLMessageDispatcher::callbackPolicy());
	_communicationManager->addMessageListener(*_clResponseDispatcher,
	        comm::CLResponseDispatcher::callbackPolicy());

	CommunicationManagerImpl::start();
}

void HostCommunicationManagerImpl::stop() {
    _communicationManager->removeMessageListener(*_clMessageDispatcher);
	_communicationManager->removeMessageListener(*_clResponseDispatcher);

	CommunicationManagerImpl::stop();
}

dcl::CLObjectRegistry& HostCommunicationManagerImpl::objectRegistry() {
    return _objectRegistry;
}

dcl::ComputeNode * HostCommunicationManagerImpl::createComputeNode(
        const std::string& url) {
    std::vector<dcl::ComputeNode *> computeNodes;

    createComputeNodes(std::vector<std::string>(1, url), computeNodes);
    if (computeNodes.empty()) throw dcl::ConnectionException("Invalid URL");
    assert(computeNodes.size() == 1);

    return computeNodes[0];
}

void HostCommunicationManagerImpl::createComputeNodes(
        const std::vector<std::string>& urls,
        std::vector<dcl::ComputeNode *>& computeNodes) {
    std::set<const gcf::Process *> processes;
    std::vector<ComputeNodeImpl *> createdComputeNodes;

    /*
     * Create process list
     */
    for (auto url : urls) {
        std::string host;
        gcf_uint16 port = DEFAULT_COMPUTE_NODE_PORT;

        resolveURL(url, host, port);

        if (host.empty()) {
            dcl::util::Logger << dcl::util::Warning
                    << "Invalid URL '" << url << '\'' << std::endl;
        } else {
            /* WARNING: Broken GCF API design
             * For the following line to work correctly,
             * USE_PROCESSLOOKUP_SINGLETON must be (un)defined, if it has been
             * (un)defined during the compilation of GCF */
            gcf::Process& process = _communicationManager->getProcessLookup().createProcess(host, port);
            if (!processes.insert(&process).second) {
                /* process already in set */
                dcl::util::Logger << dcl::util::Warning
                        << "Redundant URL '" << url << "' specified" << std::endl;
            }
        }
    }

    createComputeNodes(processes, createdComputeNodes);

    /* Automatically connect to created compute nodes
     * This API method is used by a dOpenCL ICD, which by no means will connect
     * to these compute nodes. */
    /* TODO Connect asynchronously */
    connectComputeNodes(createdComputeNodes);

    /* Resize vector, overwrite old elements and cast new elements */
    computeNodes.assign(std::begin(createdComputeNodes), std::end(createdComputeNodes));
}

void HostCommunicationManagerImpl::destroyComputeNode(dcl::ComputeNode *computeNode) {
    destroyComputeNode(dynamic_cast<ComputeNodeImpl *>(computeNode));
}

void HostCommunicationManagerImpl::createComputeNodes(
        const std::set<const gcf::Process *>& processes,
        std::vector<ComputeNodeImpl *>& computeNodes) {
    std::lock_guard<std::recursive_mutex> lock(_connectionsMutex);

    computeNodes.clear();
    computeNodes.reserve(processes.size());

    /*
     * (Lookup or) create compute nodes and add to list
     */
    for (auto process : processes) {
        assert(process != nullptr); // process must not be NULL
        /* A host process must not be specified as compute node to connect to.
         * This cannot be checked, as uninitialized processes (such as the host's
         * own process) cannot be compared to others (see implementation of
         * gcf::Process). */

        auto i = _computeNodes.find(process);
        if (i == std::end(_computeNodes)) {
            ComputeNodeImpl *computeNode = new ComputeNodeImpl(
                    *_communicationManager, *process);

            /* Process will not have a valid process ID before it is connected.
             * Hence the process ID cannot be used to identify a process. */
            _computeNodes.insert(std::make_pair(process, computeNode));
            computeNodes.push_back(computeNode);
        } else {
            computeNodes.push_back(i->second);
        }
    }
}

void HostCommunicationManagerImpl::destroyComputeNode(ComputeNodeImpl *computeNode) {
    if (computeNode) {
        const gcf::Process *process = &computeNode->process();

        computeNode->disconnect();

        {
            std::lock_guard<std::recursive_mutex> lock(_connectionsMutex);
            _computeNodes.erase(process);
        }

        delete computeNode;
        _communicationManager->getProcessLookup().freeProcess(process);
    } else {
        throw dcl::InvalidArgument(DCL_INVALID_NODE);
    }
}

void HostCommunicationManagerImpl::connectComputeNodes(
        const std::vector<ComputeNodeImpl *>& computeNodes) {
    /* Connect to compute nodes (parallelized operation) */
    ComputeNodeImpl::connect(computeNodes, ProcessImpl::Type::HOST);
}

bool HostCommunicationManagerImpl::computeNodeConnectionLost(
        const gcf::Process& process) {
    ComputeNodeImpl *computeNode = getComputeNode(process);
    if (computeNode) {
        dcl::util::Logger << dcl::util::Warning
                << "Lost connection to compute node '" << computeNode->url() << '\''
                << std::endl;

        /* TODO Attempt to re-established failed connection transparently */

        computeNode->setConnectionStatus(ComputeNodeImpl::ConnectionStatus::DISCONNECTED); // update compute node connection status
        /* FIXME Enable GCF DataStreamProcessor to detect closed streams
         * Deleting the data stream is a work-around to ensure that a closed
         * stream is removed from the GCF DataStreamProcessor. Otherwise, the
         * stream will make the DataStreamProcessor to spin in its data receipt
         * procedure, as a closed socket is indicated as incoming data. */
        computeNode->setDataStream(nullptr);
    }

    return computeNode != nullptr;
}

} /* namespace dclgcf */
