/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file ProcessImpl.h
 *
 * @date 2012-03-18
 * @author Philipp Kegel
 */

#ifndef PROCESSIMPL_H_
#define PROCESSIMPL_H_

#include <dcl/DataTransfer.h>
#include <dcl/DCLTypes.h>
#include <dcl/Process.h>
#include <dcl/Remote.h>

#include "comm/DataDispatcher.h"

#include <gcf/CommunicationManager.h>
#include <gcf/Process.h>
#include <gcf/Serializable.h>

#include <gcf/transport/IODataStream.h>

#include <chrono>
#include <condition_variable>
#include <cstddef>
#include <memory>
#include <mutex>
#include <string>

namespace dclgcf {

/**
 * @brief An implementation the dOpenCL Process interface.
 *
 * Basically, this class is a wrapper for GCF process with extensions for data
 * transfer via GCF data streams.
 */
class ProcessImpl :
public virtual dcl::Process, // implements Process
public dcl::Remote {         // extends Remote
public:
    enum class Type {
        HOST,
        COMPUTE_NODE
    };

    /* TODO Define more process connection statuses
     * DISCONNECTED
     * AUTHENTICATING
     * PROCESS_CONNECTED
     * DATASTREAM_CONNECTED
     * CONNECTED (= PROCESS_CONNECTED|DATASTREAM_CONNECTED)*/
    enum class ConnectionStatus {
        DISCONNECTED,
        CONNECTING,
        CONNECTED
    };

    /** default response timeout is 3 seconds */
    static const std::chrono::seconds DEFAULT_RESPONSE_TIMEOUT;


    /**
     * @brief Creates a process instance.
     *
     * @param[in] communicationManager      an associated GCF communication manager
     * @param[in] process                   an associated GCF process
     */
    ProcessImpl(
            gcf::CommunicationManager&  communicationManager,
            const gcf::Process&         process);
    virtual ~ProcessImpl();

    /**
     * @brief Disconnects the process
     *
     * A connection is disconnected in two steps:
     * 1. destroy data stream
     * 2. disconnect GCF process
     */
    void disconnect();
    /**
     * @brief Tests, if the compute node is connected
     *
     * @return @c true, if the compute node is connected, otherwise @c false
     */
    bool isConnected();

    const std::string& url() const;

    void sendMessage(
            gcf::Serializable& message) const;

    std::shared_ptr<dcl::DataTransfer> sendData(
            size_t          size,
            const void *    ptr);
    std::shared_ptr<dcl::DataTransfer> receiveData(
            size_t  size,
            void *  ptr);

    /**
     * @brief Returns the wrapped GCF process.
     *
     * @return a process
     */
    const gcf::Process& process() const;

    /**
     * @brief Sets the compute node's connection status.
     * A thread waiting in ComputeNode::waitForConnectionStatus will be notified.
     *
     * This method is called internally by the communication manager.
     *
     * @param[in]  status   the connection status to wait for
     */
    void setConnectionStatus(
            ConnectionStatus status);

    /**
     * @brief (Un)sets the compute nodes data stream
     *
     * This method is called internally by the communication manager.
     *
     * @param[in]  dataStream   a connected data stream or NULL
     */
    void setDataStream(
            gcf::IODataStream *dataStream);

protected:
    /**
     * @brief Waits until a given connection status has been reached, or a specified timeout expired
     *
     * @param[in] status    the connection status to wait for
     * @param[in] timeout   the maximum duration to wait for the specified connection status
     * @return @c false if timeout has been reached, otherwise @c true.
     */
    template<class Rep, class Period>
    bool awaitConnectionStatus(
            ConnectionStatus status,
            const std::chrono::duration<Rep, Period>& timeout) {
        std::lock_guard<std::recursive_mutex> lock(_connectionStatusMutex);
        while (_connectionStatus != status) {
            if (!_connectionStatusChanged.wait_for(_connectionStatusMutex, timeout) == std::cv_status::timeout) {
                /* timeout expired */
                break;
            }
        }
        return (_connectionStatus == status);
    }

    /**
     * @brief Waits until a given connection status or time point has been reached
     *
     * @param[in] status    the connection status to wait for
     * @param[in] deadline  the time to stop waiting
     * @return @c false if deadline has been reached, otherwise @c true.
     */
    template<class Clock, class Duration>
    bool awaitConnectionStatus(
            ConnectionStatus status,
            const std::chrono::time_point<Clock, Duration>& deadline) {
        std::lock_guard<std::recursive_mutex> lock(_connectionStatusMutex);
        while (_connectionStatus != status) {
            if (_connectionStatusChanged.wait_until(_connectionStatusMutex, deadline) == std::cv_status::timeout) {
                /* deadline reached */
                break;
            }
        }
        return (_connectionStatus == status);
    }

    /**
     * @brief Returns the process' associated data stream
     * This method blocks until a data stream is available
     *
     * @return a data stream
     */
    gcf::IODataStream& dataStream();

    gcf::CommunicationManager& _communicationManager;
    const gcf::Process& _process;
    gcf::IODataStream *_dataStream;
    std::condition_variable_any _dataStreamReady;

    ConnectionStatus _connectionStatus; /**< Process connection status */
    std::recursive_mutex _connectionStatusMutex; /**< Protects _connectionStatus and _dataStream */
    std::condition_variable_any _connectionStatusChanged;

    /* TODO Replace per-process data dispatcher by global data dispatcher *or*
     *      use asynchronous notification mechanism in GCF */
    comm::DataDispatcher _dataDispatcher;

private:
    /* Processes must be non-copyable */
    ProcessImpl(
            const ProcessImpl&) = delete;
    ProcessImpl& operator=(
            const ProcessImpl&) = delete;

    mutable std::string _url; /**< Process URL */
};

} /* namespace dclgcf */

#endif /* PROCESSIMPL_H_ */
