/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file Event.cpp
 *
 * Event-related unit tests
 *
 * @date 2013-02-18
 * @author Philipp Kegel
 */

#include <cstddef>
#include <cstring>
#include <iostream>
#include <vector>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE Event
#include <boost/test/unit_test.hpp>

/**
 * @brief Returns the first platform available
 *
 * @return a platform
 */
cl_platform_id getPlatform() {
    cl_platform_id platform;
    cl_uint num_platforms;
    size_t param_value_size;
    cl_int err;

    err = clGetPlatformIDs(1, &platform, &num_platforms);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_REQUIRE_MESSAGE(num_platforms >= 1, "No platform found");

    /* platform should be 'dOpenCL' */
    err = clGetPlatformInfo(platform, CL_PLATFORM_NAME, 0, NULL,
            &param_value_size);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    char platform_name[param_value_size];
    err = clGetPlatformInfo(platform, CL_PLATFORM_NAME, param_value_size,
            platform_name, NULL);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_CHECK_MESSAGE(strcmp(platform_name, "dOpenCL") == 0, "First platform must be 'dOpenCL'");

    return platform;
}

/**
 * @brief Returns the platform's first device.
 *
 * @param[in]  platform the platform
 * @return the platform's first device
 */
cl_device_id getDevice(cl_platform_id platform) {
    cl_device_id device = NULL;
    cl_uint num_devices;
    cl_int err;

    err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_ALL, 1, &device,
            &num_devices);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_REQUIRE_GE(num_devices, 1);

    return device;
}

cl_context createContext(cl_uint num_devices, cl_device_id *devices) {
    cl_context context;
    cl_int err;

    BOOST_REQUIRE_GE(num_devices, 1);
    BOOST_REQUIRE(devices != NULL);

    context = clCreateContext(0, num_devices, devices, 0, 0, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    return context;
}

cl_command_queue createCommandQueue(cl_context context, cl_device_id device) {
    cl_command_queue commandQueue;
    cl_int err;

    commandQueue = clCreateCommandQueue(context, device, 0, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    return commandQueue;
}

/******************************************************************************
 * Shared test suite fixture
 ******************************************************************************/

struct Context {
    cl_context context;
    cl_command_queue commandQueue;

    Context() {
        cl_platform_id platform = getPlatform();
        cl_device_id device = getDevice(platform);

        context = createContext(1, &device);
        commandQueue = createCommandQueue(context, device);

        BOOST_TEST_MESSAGE("Set up fixture");
    }

    ~Context() {
        cl_int err;

        /* clean up */
        err = clReleaseCommandQueue(commandQueue);
        BOOST_CHECK_EQUAL(err, CL_SUCCESS);
        err = clReleaseContext(context);
        BOOST_CHECK_EQUAL(err, CL_SUCCESS);

        BOOST_TEST_MESSAGE("Teared down fixture");
    }
};

BOOST_FIXTURE_TEST_SUITE( Event, Context )

/******************************************************************************
 * Test cases
 ******************************************************************************/

static void eventCallback(cl_event event, cl_int execution_status,
        void *user_data) {
    std::cout << "Called event callback with status " << execution_status
            << std::endl;
}

BOOST_AUTO_TEST_CASE( CreateUserEvent )
{
    cl_event event;
    cl_int err;

    event = clCreateUserEvent(context, &err);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);

    /* clean up */
    err = clReleaseEvent(event);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
}

BOOST_AUTO_TEST_CASE( GetUserEventInfo )
{
    cl_event event;
    cl_context event_context;
    cl_command_queue event_queue;
    cl_command_type command_type;
    cl_int execution_status;
    cl_int err;

    event = clCreateUserEvent(context, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    /* validate user event infos */
    err = clGetEventInfo(event, CL_EVENT_CONTEXT,
            sizeof(cl_context), &event_context, NULL);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_CHECK_EQUAL(event_context, context);

    err = clGetEventInfo(event, CL_EVENT_COMMAND_QUEUE,
            sizeof(cl_command_queue), &event_queue, NULL);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_CHECK_EQUAL(event_queue, reinterpret_cast<cl_command_queue>(0));

    err = clGetEventInfo(event, CL_EVENT_COMMAND_TYPE,
            sizeof(cl_command_type), &command_type, NULL);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_CHECK_EQUAL(command_type, CL_COMMAND_USER);

    err = clGetEventInfo(event, CL_EVENT_COMMAND_EXECUTION_STATUS,
            sizeof(cl_int), &execution_status, NULL);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_CHECK_EQUAL(execution_status, CL_SUBMITTED);

    /* clean up */
    err = clReleaseEvent(event);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
}

BOOST_AUTO_TEST_CASE( SetUserEventStatus )
{
    cl_event event;
    cl_int execution_status;
    cl_int err;

    event = clCreateUserEvent(context, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    err = clSetUserEventStatus(event, CL_COMPLETE);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    err = clGetEventInfo(event, CL_EVENT_COMMAND_EXECUTION_STATUS,
            sizeof(cl_int), &execution_status, NULL);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_CHECK_EQUAL(execution_status, CL_COMPLETE);

    /* clean up */
    err = clReleaseEvent(event);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
}

BOOST_AUTO_TEST_CASE( Callback )
{
    const size_t SIZE = 1024;
    cl_mem buffer;
    std::vector<cl_int> vec(SIZE, 0);
    cl_event start, upload;
    cl_int err;

    /* create buffer */
    buffer = clCreateBuffer(context, CL_MEM_READ_WRITE, SIZE * sizeof(cl_int), NULL, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    /* create user event */
    start = clCreateUserEvent(context, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    /* upload data to device */
    err = clEnqueueWriteBuffer(
            commandQueue,
            buffer,
            CL_FALSE,
            0, SIZE * sizeof(cl_int), &vec.front(),
            1, &start, &upload);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    /* register event callbacks */
    err = clSetEventCallback(start, CL_COMPLETE, &eventCallback, NULL);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
    err = clSetEventCallback(upload, CL_COMPLETE, &eventCallback, NULL);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);

    /* start upload */
    err = clSetUserEventStatus(start, CL_COMPLETE);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
    err = clFlush(commandQueue);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);

//    sleep(2);

    /* finish command queue */
    err = clFinish(commandQueue);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);

    /* clean up */
    err = clReleaseEvent(start);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
    err = clReleaseEvent(upload);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
    err = clReleaseMemObject(buffer);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
}

BOOST_AUTO_TEST_SUITE_END()
